#! /usr/bin/python

import os
import web
import time
import tempita
import cgitb

cgitb.enable()

# =====================================
# Variables & Constant
# =====================================

application_url = "http://www.example.com/janken"

urls = (
    '/', 'index',
    '/entry/(.*)', 'entry',
    '/admin/(.*)', 'admin'
)

loser_conditions = {
   (0,): [],
   (1,): [],
   (2,): [],
   (0, 1): [1],
   (1, 2): [2],
   (0, 2): [0],
   (0, 1, 2): []
}

ENTRY_KEY_SPLITTER = 3456789
ADMIN_KEY_SPLITTER = 1234567

DATA_FILE_NAME = "data.txt"

# =====================================
# Helper Function
# =====================================

def dprint(a):
    import pprint
    log = file("log.txt", "a")
    pprint.pprint(a, log)


# =====================================
# Model
# =====================================

class LockablePickleFile(object):
    def __init__(self, filename):
        self.file_obj = None
        self.filename = filename
        self.content = None
        self._oldcontent = None

    def read(self, default=None):
        import pickle
        if os.path.exists(self.filename):
            try:
                file_content = file(self.filename).read()
                self.content = pickle.loads(file_content)
                self._oldcontent = pickle.loads(file_content)
            except:
                self.content = default
                self._oldcontent = default
        else:
            self.content = default
            self._oldcontent = default

    def lock_and_read(self, default=None):
        self.read(default)
        self.file_obj = file(self.filename, "w")
        self._lock(self.file_obj, LockablePickleFile.LOCK_EX)

    def write(self):
        import pickle
        if self.file_obj is None:
            raise RuntimeError("This file is not locked")
        pickle.dump(self.content, self.file_obj)
        self._unlock(self.file_obj)
        self.file_obj = None

    def teardown(self):
        if self.file_obj is None:
            return
        self.content = self._oldcontent
        self.write()

    def __del__(self):
        self.teardown()

    if os.name == 'nt':
        from win32con import LOCKFILE_EXCLUSIVE_LOCK as LOCK_EX
        LOCK_SH = 0
        from win32con import LOCKFILE_FAIL_IMMEDIATELY as LOCK_NB
        
        @staticmethod
        def _lock(file, flags):
            import win32file, pywintypes
            overlapped = pywintypes.OVERLAPPED()
            hfile = win32file._get_osfhandle(file.fileno())
            win32file.LockFileEx(hfile, flags, 0, -0x10000, overlapped)

        @staticmethod
        def _unlock(file):
            import win32file, pywintypes
            overlapped = pywintypes.OVERLAPPED()
            hfile = win32file._get_osfhandle(file.fileno())
            win32file.UnlockFileEx(hfile, 0, -0x10000, overlapped)

    elif os.name == 'posix':
        from fcntl import LOCK_EX, LOCK_SH, LOCK_NB

        @staticmethod
        def _lock(file, flags):
            import fcntl
            fcntl.flock(file.fileno(), flags)

        @staticmethod
        def _unlock(file):
            import fcntl
            fcntl.flock(file.fileno(), fcntl.LOCK_UN)
    else:
        raise RuntimeError()


def get_id():
    import random
    return random.randint(0, 500) * 1500000 + int(time.time())


def create_admin_key(id):
    top = id / ADMIN_KEY_SPLITTER
    rest = id % ADMIN_KEY_SPLITTER
    return ("ad,%d,%d" % (rest, top)).encode("base64").strip("\n")


def create_entry_key(id):
    top = id / ENTRY_KEY_SPLITTER
    rest = id % ENTRY_KEY_SPLITTER
    return ("en,%d,%d" % (rest, top)).encode("base64").strip("\n")


def reverse_from_key(key):
    head, rest, top = key.decode("base64").split(",")
    if head == "ad":
        return ADMIN_KEY_SPLITTER * int(top) + int(rest)
    else:
        return ENTRY_KEY_SPLITTER * int(top) + int(rest)


class Entry(object):
    def __init__(self, name, hands):
        self.name = name
        self.hands = hands

    def get_hand(self, round_number):
        round_number = round_number % len(self.hands)
        return int(self.hands[round_number])

    def is_lose(self, round_number, loser_hand):
        return self.get_hand(round_number) in loser_hand


# =====================================
# Logic
# =====================================


def janken(players, round_number):
    draw_round_number = 0
    result = {"winner":[], "loser":[], "draw":[], "round":0}
    while True:
        field = set([player.get_hand(round_number) for player in players])
        hands = [(player.name, player.get_hand(round_number))
                    for player in players]
        loser = loser_conditions[tuple(sorted(field))]
        if len(loser) != 0:
            result["round"] = round_number
            for player in players:
                if player.is_lose(round_number, loser):
                    result["loser"].append(player)
                else:
                    result["winner"].append(player)
            return result
        draw_round_number += 1
        round_number += 1
        if draw_round_number > 9:
            result["round"] = round_number
            result["draw"] = players
            return result


class ranked_list(object):
    def __init__(self, items):
        self.rank = {1:items}

    def get_duplicate_rank_items(self, start_number = -1):
        for i, rank_number in enumerate(sorted(self.rank.iterkeys())):
            if start_number > i:
                continue
            if len(self.rank[rank_number]) != 1:
                return rank_number, self.rank[rank_number]
        return None, None

    def remove(self, rank_number):
        if rank_number in self.rank:
            del self.rank[rank_number]

    def add(self, rank_number, items):
        self.rank[rank_number] = items

    def get_list(self):
        result = []
        for rank_number in sorted(self.rank.iterkeys()):
            result.append((rank_number, self.rank[rank_number]))
        return result


def get_result(entries):
    entries = ranked_list(entries)
    round_number = 0
    start_number = -1

    while True:
        rank, players = entries.get_duplicate_rank_items(start_number)
        if rank is None:
            return entries.get_list()
        result = janken(players, round_number)
        if len(result["draw"]) != 0:
            start_number = rank + 1
            continue
        round_number = result["round"]
        entries.remove(rank)
        entries.add(rank, result["winner"])
        entries.add(rank+len(result["winner"]), result["loser"])
        round_number += 1


def is_data_valid(data_file, entry_id):
    if entry_id not in data_file.content:
        return False
    if time.time() > data_file.content[entry_id]["erase_time"]:
        new_data_file = LockablePickleFile(data_file.filename)
        new_data_file.lock_and_read(default={})
        data = data_file.content[entry_id]
        if data["close"]:
            del new_data_file.content[entry_id]
        else:
            data["close"] = True
            data["erase_time"] = time.time() + 60*60*24*7
        new_data_file.write()
        return False
    return True


# =====================================
# Controllers
# =====================================


class index:
    def GET(self):
        web.header("Content-Type", "text/html; charset=UTF-8")
        print file("index.html").read()

    def POST(self):
        web.header("Content-Type", "text/html; charset=utf-8")
        form = web.input()
        
        try:
            desc = form.desc
            id = get_id()
            
            data_file = LockablePickleFile(DATA_FILE_NAME)
            data_file.lock_and_read(default={})
            deadline_int = time.time() + 60*60*24*7
            data_file.content[id] = {"desc":desc,
                                     "close":False,
                                     "entries":[],
                                     "erase_time":deadline_int
                                    }
            admin_url = "%s/admin/%s" % (application_url, create_admin_key(id))
            entry_url = "%s/entry/%s" % (application_url, create_entry_key(id))
            deadline = time.strftime("%Y/%m/%d %H:%M",
                          time.localtime(deadline_int))
            template = tempita.HTMLTemplate.from_filename("end_create.html")
            print template.substitute(locals())
            data_file.write()
        except ValueError:
            print file("index.html").read()


def show_result(data):
    janken_result = get_result(data["entries"])
    result_list = []
    for rank, players in janken_result:
        names = [player.name for player in players]
        result_list.append(tempita.bunch(rank=rank, names=names))
    deadline = time.strftime("%Y/%m/%d %H:%M",
                      time.localtime(data["erase_time"]))
    template = tempita.HTMLTemplate.from_filename("result.html")
    print template.substitute(locals())


def show_error():
    print file("error.html").read()


class entry:
    def GET(self, key):
        web.header("Content-Type", "text/html; charset=utf-8")
        data_file = LockablePickleFile(DATA_FILE_NAME)
        id = reverse_from_key(key)
        data_file.read(default={})
        if not is_data_valid(data_file, id):
            show_error()
            return
        data = data_file.content[id]
        if data["close"]:
            return show_result(data)
        desc = data["desc"]
        template = tempita.HTMLTemplate.from_filename("entry.html")
        print template.substitute(locals())

    def POST(self, key):
        web.header("Content-Type", "text/html; charset=utf-8")
        form = web.input()
        name = form.name
        if len(name) == 0:
            self.GET(key)
            return

        hands = "".join([form["hand%d" % i] for i in xrange(10)])

        id = reverse_from_key(key)
        try:
            data_file = LockablePickleFile(DATA_FILE_NAME)
            data_file.lock_and_read()
            data = data_file.content[id]
            if not data["close"]:
                entry_number = len(data["entries"]) + 1
                deadline = time.strftime("%Y/%m/%d %H:%M",
                              time.localtime(data["erase_time"]))
                template = tempita.HTMLTemplate.from_filename("end_entry.html")
                print template.substitute(locals())
                data["entries"].append(Entry(name, hands))
                data_file.write()
            else:
                show_result(data)
        except KeyError:
            show_error()


class admin:
    def GET(self, key):
        web.header("Content-Type", "text/html; charset=utf-8")

        id = reverse_from_key(key)
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.read()
        if not is_data_valid(data_file, id):
            show_error()
            return
        data = data_file.content[id]

        if not data["close"]:
            names = [entry.name for entry in data["entries"]]
            entry_number = len(data["entries"])
            deadline = time.strftime("%Y/%m/%d %H:%M",
                          time.localtime(data["erase_time"]))
            template = tempita.HTMLTemplate.from_filename("admin.html")
            print template.substitute(locals())
        else:
            self.show_end_html(key, data)

    def POST(self, key):
        web.header("Content-Type", "text/html; charset=utf-8")
        id = reverse_from_key(key)
        data_file = LockablePickleFile(DATA_FILE_NAME)
        data_file.lock_and_read()
        try:
            data = data_file.content[id]
            data["close"] = True
            data["erase_time"] = time.time() + 60*60*24*7
            self.show_end_html(key, data)
            data_file.write()
        except KeyError:
            show_error()

    def show_end_html(self, key, data):
        id = reverse_from_key(key)
        entry_url = "%s/entry/%s" % (application_url, create_entry_key(id))
        deadline = time.strftime("%Y/%m/%d %H:%M",
                          time.localtime(data["erase_time"]))
        template = tempita.HTMLTemplate.from_filename("end_admin.html")
        print template.substitute(locals())


if __name__ == "__main__":
    web.webapi.internalerror = web.debugerror
    web.run(urls, globals())
